/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/param.h>

#include <machine/limits.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bigint.h"

#define BIGINT_WORDSIZE	(sizeof(uint64_t) * CHAR_BIT)

#define BIGINT_NUM_WORDS(x)	((x) / BIGINT_WORDSIZE)
#define DIGIT64(x, i, n)	(((i) < (n)) ? (x)->un.v8[i] : 0)
#define DIGIT32(x, i, u)	(((i) >= 2 * (u)) ? 0 : (x)->un.v4[i])

#define HIGH32(x)		(((x) & 0xffffffff00000000) >> 32)
#define LOW32(x)		((uint32_t)(x) & 0xffffffff)
#define MAX32			0xffffffffUL

#ifndef USE_ARG
#define USE_ARG(x)		/*LINTED*/(void)&(x)
#endif

#ifndef __arraycount
#define	__arraycount(__x)	(sizeof(__x) / sizeof(__x[0]))
#endif

#ifndef howmany
#define	howmany(x, y)		(((x)+((y)-1))/(y))
#endif

#define IS_EVEN(x)		(((x)->un.v1[0] % 2) == 0)

/* return the 32-bit number corresponding to the string buffer */
static uint32_t
getword(const char *str, const size_t len, const size_t base, uint64_t *mult)
{
	size_t	size;
	size_t	i;
	char	buf[8 + 1];

	(void) memset(buf, 0x0, sizeof(buf));
	size = (len > 8) ? 8 : len;
	strlcpy(buf, str, size + 1);
	if (mult) {
		for (*mult = 1, i = 0 ; i < size ; i++, *mult *= base) {
		}
	}
	return (uint32_t)strtoul(buf, NULL, (int)base);
}

static inline size_t
getwindowsize(size_t bits)
{
	return (bits > 671) ? 6 :
		(bits > 239) ? 5 :
		(bits > 79) ? 4 :
		(bits > 23) ? 3 : 1;
}

/*****************************************************************************/

int
bigint_word_arith(uint4096_t *res, const uint4096_t *a, char op, const uint64_t w, const size_t nbits)
{
	uint4096_t	tmp;
	size_t		words;
	size_t		i;

	switch(op) {
	case '=':
		(void) memset(res, 0x0, sizeof(*res));
		res->un.v8[0] = w;
		return 1;
	case '+':
	case '-':
		if (w == 0) {
			if (res != a) {
				(void) memcpy(res, a, sizeof(*res));
			}
		} else {
			(void) memset(&tmp, 0x0, sizeof(tmp));
			tmp.un.v8[0] = w;
			if (op == '+') {
				bigint_add(res, a, &tmp, nbits);
			} else {
				bigint_sub(res, a, &tmp, nbits);
			}
		}
		return 1;
	case '*':
		if (w == 0) {
			(void) memset(res, 0x0, sizeof(*res));
		} else if (w == 1) {
			if (res != a) {
				(void) memcpy(res, a, sizeof(*res));
			}
		} else {
			(void) memset(&tmp, 0x0, sizeof(tmp));
			tmp.un.v8[0] = w;
			bigint_mul(res, a, &tmp, nbits);
		}
		return 1;
	case '/':
	case '%':
		if (w == 1) {
			if (res != a) {
				(void) memcpy(res, a, sizeof(*res));
			}
		} else if (w == 0) {
			return 0;
		} else {
			(void) memset(&tmp, 0x0, sizeof(tmp));
			tmp.un.v8[0] = w;
			if (op == '/') {
				bigint_div(res, a, &tmp, nbits);
			} else {
				bigint_mod(res, a, &tmp, nbits);
			}
		}
		return 1;
        case '?':
                words = BIGINT_NUM_WORDS(nbits);
                for (i = words - 1 ; i > 0 ; --i) { 
                        if (a->un.v8[i] != 0) {
                                return 0;
                        }
                }
                return (a->un.v8[0] == w);
	default:
		return 0;
	}
}

/*************************************************************************/

/* add */
int
bigint_add(uint4096_t *res, const uint4096_t *a, const uint4096_t *b, const size_t nbits)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	lo;
	size_t		words;
	size_t		i;

	if (bigint_iszero(a, nbits)) {
		(void) memcpy(&ans, b, sizeof(ans));
	} else if (bigint_iszero(b, nbits)) {
		(void) memcpy(&ans, a, sizeof(ans));
	} else {
		words = BIGINT_NUM_WORDS(nbits) * 2;
		(void) memset(&ans, 0x0, sizeof(ans));
		for (carry = 0, i = 0 ; i < words ; i++) {
			lo = carry + b->un.v4[i] + a->un.v4[i];
			ans.un.v4[i] = LOW32(lo);
			carry = LOW32(lo >> 32);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

/* subtract */
int
bigint_sub(uint4096_t *res, const uint4096_t *a, const uint4096_t *b, const size_t nbits)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	m;
	size_t		words;
	size_t		i;

	if (bigint_iszero(a, nbits)) {
		(void) memcpy(&ans, b, sizeof(ans));
	} else if (bigint_iszero(b, nbits)) {
		(void) memcpy(&ans, a, sizeof(ans));
	} else {
		words = BIGINT_NUM_WORDS(nbits) * 2;
		(void) memset(&ans, 0x0, sizeof(ans));
		for (carry = 0, i = 0 ; i < words ; i++) {
			m = (uint64_t)a->un.v4[i] - DIGIT32(b, i, words) + carry;
			carry = (m > MAX32);
			ans.un.v4[i] = LOW32(m);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

/* division/modulo */
int
bigint_divmod(uint4096_t *quot, uint4096_t *rem, const uint4096_t *x, const uint4096_t *y, const size_t nbits)
{
	uint4096_t	q;
	uint4096_t	r;
	uint4096_t	d;
	size_t		xdigits;
	size_t		ydigits;
	size_t		i;

	if (bigint_iszero(y, nbits)) {
		return 0;
	}
	(void) memset(&q, 0x0, sizeof(q));
	if (bigint_iszero(x, nbits) || bigint_cmp(x, y, nbits) < 0) {
		(void) memcpy(&r, x, sizeof(r));
	} else {
		xdigits = bigint_numbits(x, nbits);
		ydigits = bigint_numbits(y, nbits);
		(void) memcpy(&d, y, sizeof(d));
		(void) memcpy(&r, x, sizeof(r));
		bigint_lshift(&d, &d, xdigits - ydigits, nbits);
		for (i = xdigits - ydigits ; (ssize_t)i >= 0 ; --i) {
			bigint_lshift(&q, &q, 1, nbits);
			if (bigint_cmp(&r, &d, nbits) >= 0) {
				q.un.v1[0] |= 1;
				bigint_sub(&r, &r, &d, nbits);
			}
			bigint_rshift(&d, &d, 1, nbits);
		}
	}
	if (quot) {
		(void) memcpy(quot, &q, sizeof(*quot));
	}
	if (rem) {
		(void) memcpy(rem, &r, sizeof(*rem));
	}
	return 1;
}

/* modulo */
int
bigint_mod(uint4096_t *r, const uint4096_t *x, const uint4096_t *y, size_t nbits)
{
	return bigint_divmod(NULL, r, x, y, nbits);
}

/* divide operation */
int
bigint_div(uint4096_t *q, const uint4096_t *x, const uint4096_t *y, size_t nbits)
{
	return bigint_divmod(q, NULL, x, y, nbits);
}

/* prod = a * b */
int
bigint_mul(uint4096_t *prod, const uint4096_t *a, const uint4096_t *b, const size_t nbits)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	prodlo;
	uint64_t	prodhi;
	uint64_t	loword;
	uint64_t	adigit;
	size_t		words;
	size_t		i;
	size_t		j;

	(void) memset(&ans, 0x0, sizeof(ans));
	if (!bigint_iszero(a, nbits) && !bigint_iszero(b, nbits)) {
		words = BIGINT_NUM_WORDS(nbits) * 2;
		for (j = 0; j < words; j++) {
			adigit = DIGIT32(a, j, words);
			for (carry = 0, i = 0; i < words; i++) {
				prodlo = adigit * DIGIT32(b, i, words);
				prodhi = HIGH32(prodlo);
				prodlo = LOW32(prodlo);
				loword = LOW32(ans.un.v4[i + j]) + carry;
				if ((carry = (loword > MAX32)) == 1) { 
					loword &= MAX32;
				}
				if ((loword += prodlo) > MAX32) {
					carry += 1;
					loword &= MAX32;
				}
				carry += prodhi;
				ans.un.v4[i + j] = LOW32(loword);
			}
			if (carry) {
				ans.un.v4[i + j] += LOW32(carry);
			}
		}
	}
	(void) memcpy(prod, &ans, sizeof(*prod));
	return 1;
}

/* wrapper function for arith functions */
int
bigint_arith(uint4096_t *res, const uint4096_t *a, char op, const uint4096_t *b, const size_t nbits)
{
	switch(op) {
	case '+':
		return bigint_add(res, a, b, nbits);
	case '-':
		return bigint_sub(res, a, b, nbits);
	case '*':
		return bigint_mul(res, a, b, nbits);
	case '/':
		return bigint_divmod(res, NULL, a, b, nbits);
	case '%':
		return bigint_divmod(NULL, res, a, b, nbits);
	default:
		return 0;
	}
}

/* compare abs(b) and abs(a) */
int
bigint_cmp(const uint4096_t *a, const uint4096_t *b, const size_t nbits)
{
	int64_t	cmp;
	size_t	words;
	size_t	i;

	words = BIGINT_NUM_WORDS(nbits);
	for (i = words - 1 ; (ssize_t)i >= 0 ; i--) {
		if ((cmp = (DIGIT64(a, i, words) - DIGIT64(b, i, words))) < 0) {
			return -1;
		} else if (cmp > 0) {
			return 1;
		}
	}
	return 0;
}

/* is abs(x) == 0 */
int
bigint_iszero(const uint4096_t *x, const size_t nbits)
{
	size_t	words;
	size_t	i;

	words = BIGINT_NUM_WORDS(nbits);
	for (i = 0; i < words; i++) {
		if (x->un.v8[i] != 0) {
			return 0;
		}
	}
	return 1;
}

/* nshift must be < BIGINT_WORDSIZE */
int
bigint_rshift(uint4096_t *res, const uint4096_t *a, const size_t nshift, const size_t nbits)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	mask;
	size_t		words;
	size_t		i;

	words = BIGINT_NUM_WORDS(nbits);
	if (nshift < BIGINT_WORDSIZE) {
		if (nshift == 0) {
			(void) memcpy(res, a, sizeof(*res));
		} else {
			(void) memset(&ans, 0x0, sizeof(ans));
			mask = (((uint64_t)1 << nshift) - 1);
			for (carry = 0, i = words - 1 ; (ssize_t)i >= 0 ; --i) {
				ans.un.v8[i] = ((a->un.v8[i] >> nshift) | carry);
				carry = (a->un.v8[i] & mask) << (BIGINT_WORDSIZE - nshift);
			}
			(void) memcpy(res, &ans, sizeof(*res));
		}
		return 1;
	}
	return 0;
}
 
/* n must be < BIGINT_WORDSIZE */
int
bigint_lshift(uint4096_t *res, const uint4096_t *a, const size_t nshift, const size_t nbits)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	mask;
	size_t		words;
	size_t		i;

	words = BIGINT_NUM_WORDS(nbits);
	if (nshift < BIGINT_WORDSIZE) {
		if (nshift == 0) {
			(void) memcpy(res, a, sizeof(*res));
		} else {
			(void) memset(&ans, 0x0, sizeof(ans));
			mask = (((uint64_t)1 << nshift) - 1) << (BIGINT_WORDSIZE - nshift);
			for (carry = 0, i = 0 ; i < words ; i++) {
				ans.un.v8[i] = ((a->un.v8[i] << nshift) | carry);
				carry = (a->un.v8[i] & mask) >> (BIGINT_WORDSIZE - nshift);
			}
			(void) memcpy(res, &ans, sizeof(*res));
		}
		return 1;
	}
	return 0;
}
 
/* print to stream */
size_t
bigint_fprintf(FILE *fp, const uint4096_t *m, const size_t nbits)
{
	size_t	cc;
	char	buf[1024 + 1];

	cc = bigint_snprintf(buf, sizeof(buf), m, nbits);
	fprintf(fp, "%s", buf);
	return cc;
}

/* print to buffer */
size_t
bigint_snprintf(char *buf, const size_t size, const uint4096_t *m, const size_t nbits)
{
	size_t		 words;
	size_t		 off;
	size_t		 cc;
	size_t		 i;
	int		 leading;

	words = BIGINT_NUM_WORDS(nbits);
	for (off = 0, leading = 1, i = words - 1 ; (ssize_t)i >= 0 && off < size - 1 ; --i) {
		if (leading && m->un.v8[i] != 0) {
			leading = 0;
			cc = snprintf(&buf[off], size - off, "%" PRIx64, m->un.v8[i]);
		} else if (!leading) {
			cc = snprintf(&buf[off], size - off, "%016" PRIx64, m->un.v8[i]);
		} else {
			cc = 0;
		}
		off += cc;
	}
	return off;
}

/* convert a hex ASCII string to a fixed-precision integer - base is 10 or 16 */
int
bigint_str2bigint(uint4096_t *a, const char *str, const size_t len, const size_t base, const size_t nbits)
{
	uint64_t	mult;
	uint64_t	w;
	size_t		i;

	(void) memset(a, 0x0, sizeof(*a));
	for (i = 0 ; i < len ; i += 8) {
		w = getword(&str[i], len - i, base, &mult);
		bigint_word_arith(a, a, '*', mult, nbits);
		bigint_word_arith(a, a, '+', w, nbits);
	}
	return 1;
}

/* put random values into 'a' */
int
bigint_random(uint4096_t *a, const char *type, size_t nbits)
{
	size_t		words;
	size_t		last;
	size_t		size;
	size_t		i;
	int		use_arcrandom;

	memset(a, 0x0, sizeof(*a));
	words = BIGINT_NUM_WORDS(nbits);
	use_arcrandom = (strcasecmp(type, "strong") == 0);
	size = (use_arcrandom) ? 4 : sizeof(long);
	last = (size == 4) ? words + words : words;
	for (i = 0 ; i < last ; ) {
		if (use_arcrandom) {
			a->un.v4[i++] = arc4random();
			a->un.v4[i++] = arc4random();
		} else if (size == 4) {
			a->un.v4[i++] = (uint32_t)random();
			a->un.v4[i++] = (uint32_t)random();
		} else {
			a->un.v8[i++] = random();
		}
	}
	return 1;
}

/* convert a fixed precision number to a binary */
size_t
bigint_bigint2bin(const uint4096_t *a, size_t nbits, uint8_t *buf, size_t size)
{
	size_t		i;

	for (i = 0 ; i < nbits && i < size ; i++) {
		buf[size - 1 - i] = a->un.v1[i];
	}
	return i;
}

/* convert a binary to a fixed precision number */
size_t
bigint_bin2bigint(uint4096_t *a, size_t nbits, const uint8_t *buf, size_t size)
{
	size_t		i;

	for (i = 0 ; i < nbits && i < size ; i++) {
		a->un.v1[i] = buf[size - 1 - i];
	}
	return i;
}

/* find the number of significant bits in 'x' */
size_t
bigint_numbits(const uint4096_t *x, const size_t nbits)
{
	static const uint32_t	nums[] = { 1, 2, 4, 8, 16, 32, 64, 128 };
	size_t			bytes;
	size_t			i;
	size_t			j;

	bytes = nbits / 8;
	for (i = bytes - 1; (ssize_t)i >= 0 && x->un.v1[i] == 0 ; i--) {
	}
	for (j = 0 ; j < __arraycount(nums) && x->un.v1[i] >= nums[j] ; j++) {
	}
	return ((i + 1) * 8) + (j - 1);
}

/* and the number of bytes */
size_t
bigint_numbytes(const uint4096_t *x, const size_t nbits)
{
	return howmany(bigint_numbits(x, nbits), CHAR_BIT);
}
        
/* test if a bit is set */
int
bigint_bit_is_set(const uint4096_t *a, size_t bit, size_t nbits)
{
	if (bit > nbits || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] & (1 << ((unsigned)bit % CHAR_BIT));
}

/* set a bit */
int
bigint_bit_set(uint4096_t *a, size_t bit, size_t nbits)
{
	if (bit > nbits || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] |= (1 << ((unsigned)bit % CHAR_BIT));
}

/* clear a bit */
int
bigint_bit_clear(uint4096_t *a, size_t bit, size_t nbits)
{
	if (bit > nbits || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] &= ~(1 << ((unsigned)bit % CHAR_BIT));
}

/* bitwise operation - and, or, xor */
int
bigint_bit_op(uint4096_t *res, const uint4096_t *a, char op, const uint4096_t *b, const size_t nbits)
{
	size_t	words;
	size_t	i;

	words = BIGINT_NUM_WORDS(nbits);
	for (i = 0 ; i < words ; i++) {
		switch(op) {
		case '&':
			res->un.v8[i] = (a->un.v8[i] & b->un.v8[i]);
			break;
		case '|':
			res->un.v8[i] = (a->un.v8[i] | b->un.v8[i]);
			break;
		case '^':
			res->un.v8[i] = (a->un.v8[i] ^ b->un.v8[i]);
			break;
		case '1':
			res->un.v8[i] = 0xffffffffffffffffULL;
			break;
		case '0':
			res->un.v8[i] = 0ULL;
			break;
		default:
			return 0;
		}
	}
	return 1;
}

/*************************************************************************/
/* higher-level functions */
/*************************************************************************/

/* find gcd */
int
bigint_gcd(uint4096_t *gcd, const uint4096_t *xorig, const uint4096_t *yorig, const size_t nbits)
{
	uint4096_t	g;
	uint4096_t	x;
	uint4096_t	y;
	uint4096_t	t;
	uint4096_t	zero;

	(void) memcpy(&x, xorig, sizeof(x));
	(void) memcpy(&y, yorig, sizeof(y));
	(void) memset(&zero, 0x0, sizeof(zero));
	bigint_word_arith(&g, NULL, '=', 1, nbits);
	while (IS_EVEN(&x) && IS_EVEN(&y)) {
		bigint_rshift(&x, &x, 1, nbits);
		bigint_rshift(&y, &y, 1, nbits);
		bigint_lshift(&g, &g, 1, nbits);
	}
	while (bigint_cmp(&x, &zero, nbits) != 0) {
		while (IS_EVEN(&x) && bigint_cmp(&x, &zero, nbits) != 0) {
			bigint_rshift(&x, &x, 1, nbits);
		}
		while (IS_EVEN(&y) && bigint_cmp(&y, &zero, nbits) != 0) {
			bigint_rshift(&y, &y, 1, nbits);
		}
		if (bigint_cmp(&x, &y, nbits) > 0) {
			bigint_sub(&t, &x, &y, nbits);
		} else {
			bigint_sub(&t, &y, &x, nbits);
		}
		bigint_rshift(&t, &t, 1, nbits);
		if (bigint_cmp(&x, &y, nbits) >= 0) {
			(void) memcpy(&x, &t, sizeof(x));
		} else {
			(void) memcpy(&y, &t, sizeof(x));
		}
	}
	bigint_mul(gcd, &g, &y, nbits);
	return 1;
}

/* exponentiation */
int
bigint_exp(uint4096_t *res, const uint4096_t *a, const uint4096_t *p, size_t nbits)
{
	uint4096_t	ans;
	uint4096_t	v;
	size_t		bits;
	size_t		i;

	(void) memcpy(&v, a, sizeof(v));
	bits = bigint_numbits(p, nbits);
	if (IS_EVEN(p)) {
		bigint_word_arith(&ans, NULL, '=', 1, nbits);
	} else {
		(void) memcpy(&ans, a, sizeof(ans));
	}
	for (i = 1 ; i < bits ; i++) {
		bigint_mul(&v, &v, &v, nbits);
		if (bigint_bit_is_set(p, i, nbits)) {
			bigint_mul(&ans, &ans, &v, nbits);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

int
bigint_mod_inverse(uint4096_t *ret, const uint4096_t *in, const uint4096_t *a, const uint4096_t *n, const size_t nbits)
{
	uint4096_t	tmp;
	uint4096_t	A;
	uint4096_t	B;
	uint4096_t	X;
	uint4096_t	Y;
	uint4096_t	M;
	uint4096_t	D;
	uint4096_t	T;
	uint4096_t	R;

	(void) memset(&X, 0x0, sizeof(X));
	(void) memset(&Y, 0x0, sizeof(Y));
	(void) memset(&M, 0x0, sizeof(M));
	(void) memset(&D, 0x0, sizeof(D));
	(void) memset(&T, 0x0, sizeof(T));
	(void) memcpy(&R, in, sizeof(R));
	bigint_word_arith(&X, NULL, '=', 1, nbits);
	bigint_word_arith(&Y, NULL, '=', 0, nbits);
	(void) memcpy(&B, a, sizeof(B));
	(void) memcpy(&A, n, sizeof(A));
	while (!bigint_iszero(&B, nbits)) {
		if (bigint_numbits(&A, nbits) == bigint_numbits(&B, nbits)) {
			bigint_word_arith(&D, NULL, '=', 1, nbits);
			bigint_sub(&M, &A, &B, nbits);
		} else if (bigint_numbits(&A, nbits) == bigint_numbits(&B, nbits) + 1) {
			bigint_lshift(&T, &B, 1, nbits);
			if (bigint_cmp(&A, &T, nbits) < 0) {
				bigint_word_arith(&D, NULL, '=', 1, nbits);
				bigint_sub(&M, &A, &B, nbits);
			} else {
				bigint_sub(&M, &A, &T, nbits);
				bigint_add(&D, &T, &B, nbits);
				if (bigint_cmp(&A, &D, nbits) < 0) {
					bigint_word_arith(&D, NULL, '=', 2, nbits);
				} else {
					bigint_word_arith(&D, NULL, '=', 3, nbits);
					bigint_sub(&M, &M, &B, nbits);
				}
			}
		} else {
			bigint_divmod(&D, &M, &A, &B, nbits);
		}
		(void) memcpy(&tmp, &A, sizeof(tmp));
		(void) memcpy(&A, &B, sizeof(A));
		(void) memcpy(&B, &M, sizeof(B));
		if (bigint_word_arith(NULL, &D, '?', 1, nbits) == 0) {
			bigint_add(&tmp, &X, &Y, nbits);
		} else {
			if (bigint_word_arith(NULL, &D, '?', 2, nbits)) {
				bigint_lshift(&tmp, &X, 1, nbits);
			} else if (bigint_word_arith(NULL, &D, '?', 4, nbits)) {
				bigint_lshift(&tmp, &X, 2, nbits);
			} else {
				bigint_mul(&tmp, &D, &X, nbits);
			}
			bigint_add(&tmp, &tmp, &Y, nbits);
		}
		(void) memcpy(&M, &Y, sizeof(M));
		(void) memcpy(&Y, &X, sizeof(Y));
		(void) memcpy(&X, &tmp, sizeof(X));
	}
	if (bigint_word_arith(NULL, &A, '?', 1, nbits)) {
		if (bigint_cmp(&Y, n, nbits) < 0) {
			(void) memcpy(&R, &Y, sizeof(R));
		} else {
			bigint_mod(&R, &Y, n, nbits);
		}
	} else {
		fprintf(stderr, "no inverse\n");
	}
	(void) memcpy(ret, &R, sizeof(*ret));
	return 1;
}

int
bigint_mod_mul(uint4096_t *r, const uint4096_t *a, const uint4096_t *b, const uint4096_t *m, const size_t nbits)
{
	uint4096_t	t;

	(void) memset(&t, 0x0, sizeof(t));
	bigint_mul(&t, a, b, nbits);
	bigint_mod(r, &t, m, nbits);
	return 1;
}

int
bigint_mod_exp(uint4096_t *r, const uint4096_t *a, const uint4096_t *p, const uint4096_t *m, const size_t nbits)
{
	uint4096_t	val[32];
	uint4096_t	d;
	ssize_t		wstart;
	ssize_t		i;
	size_t		j;
	size_t		bits;
	size_t		wend;
	size_t		windowsize;
	size_t		wvalue;
	size_t		start;

	if ((bits = bigint_numbits(p, nbits)) == 0) {
		bigint_word_arith(r, NULL, '=', 1, nbits);
		return 0;
	}
	(void) memset(&d, 0x0, sizeof(d));
	(void) memset(&val[0], 0x0, sizeof(val[0]));
	bigint_mod(&val[0], a, m, nbits);
	if (bigint_iszero(&val[0], nbits)) {
		bigint_word_arith(r, NULL, '=', 0, nbits);
		return 1;
	}
	/* set up the window values */
	if ((windowsize = getwindowsize(bits)) > 1) {
		bigint_mod_mul(&d, &val[0], &val[0], m, nbits);
		j = ((unsigned)1 << ((unsigned)windowsize - 1));
		for (i = 1 ; i < (ssize_t)j ; i++) {
			(void) memset(&val[i], 0x0, sizeof(val[i]));
			bigint_mod_mul(&val[i], &val[i - 1], &d, m, nbits);
		}
	}
	start = 1;
	wvalue = 0;
	wstart = bits - 1;
	wend = 0;
	bigint_word_arith(r, NULL, '=', 1, nbits);
	for (;;) {
		if (!bigint_bit_is_set(p, (size_t)wstart, nbits)) {
			if (!start) {
				bigint_mod_mul(r, r, r, m, nbits);
			}
			if (wstart == 0) {
				break;
			}
			wstart -= 1;
			continue;
		}
		j = wstart;
		wvalue = 1;
		wend = 0;
		for (i = 1 ; i < (ssize_t)windowsize ; i++) {
			if (wstart - i < 0) {
				break;
			}
			if (bigint_bit_is_set(p, (size_t)(wstart - i), nbits)) {
				wvalue <<= (i - wend);
				wvalue |= 1;
				wend = i;
			}
		}
		j = wend + 1;
		if (!start) {
			for (i = 0 ; i < (ssize_t)j ; i++) {
				bigint_mod_mul(r, r, r, m, nbits);
			}
		}
		bigint_mod_mul(r, r, &val[wvalue >> 1], m, nbits);
		wstart -= wend + 1;
		wvalue = 0;
		start = 0;
		if (wstart < 0) {
			break;
		}
	}
	return 1;
}
